/*pas9782Driver.c */
/* pas9782.c,v 1.1 2006/04/21 20:22:58 xusf Exp */
/***********************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of
* California, as Operator of Los Alamos National Laboratory, and
* Berliner Elektronenspeicherring-Gesellschaft m.b.H. (BESSY).
* asynDriver is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
***********************************************************************/
/* driver support
 * Author: Shifu Xu
 * Date:   06JAN2006
 */
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include "epicsPrint.h"
#include <ctype.h>
#include <cantProceed.h>
#include <epicsStdio.h>
#include <epicsMutex.h>
#include <epicsEvent.h>
#include <epicsThread.h>
#include <epicsExport.h>
#include <iocsh.h>
#include <asynDriver.h>
#include <asynDrvUser.h>
#include <asynUInt32Digital.h>
#include <devLib.h>
#include <asynFloat64.h>
#define NCHANNELS 8
typedef volatile struct channel_reg {
    unsigned short freq0[2];
    unsigned short freq1[2];
    unsigned short phase[4];
}   channel_reg_t;
typedef volatile struct PAS9782 {
    unsigned short id[16];
    unsigned short reserved0[16];
    unsigned short csr;
    unsigned short reserved1[7];
    unsigned short ccr[8];
    unsigned short reserved2[16];
    channel_reg_t channel[8];
}   PAS9782_t;
typedef enum requestType {
    FREQUENCY0,
    PHASE0,
    FREQUENCY1,
    PHASE1,
    AMPLITUDE,
    RESET,
    ENABLE,
    FREQ_SELECT,
    PHASE_SELECT,
    CARD_STATUS,
}   requestType;
typedef struct mapType {
    requestType type;
    char *requestMsg;
}   mapType;
static mapType map[] = {
    {FREQUENCY0, "frequency0"},
    {FREQUENCY1, "frequency1"},
    {PHASE0, "phase0"},
    {PHASE1, "phase1"},
    {RESET, "reset"},
    {ENABLE, "enable"},
    {AMPLITUDE, "amplitude"},
    {FREQ_SELECT, "freq_select"},
    {PHASE_SELECT, "phase_select"},
    {CARD_STATUS, "card_status"},
};
typedef struct drvPvt {
    const char *portName;
    int connected;
    int channel_connected[NCHANNELS + 1];
    epicsMutexId lock;
    asynInterface common;
    asynInterface asynDrvUser;
    asynInterface asynUInt32Digital;
    asynInterface asynFloat64;
    PAS9782_t *pboard_register;
}   drvPvt_t;
typedef struct drvUserData {
    requestType type;
}   drvUserData;
static asynStatus create(void *drvPvt, asynUser * pasynUser,
			     const char *drvInfo, const char **pptypeName,
			     size_t * psize);
static asynStatus getType(void *ddrvPvt, asynUser * pasynUser,
			      const char **pptypeName, size_t * psize);
static asynStatus destroy(void *drvPvt, asynUser * pasynUser);
static asynDrvUser drvUser = {create, getType, destroy};
int pas9782DriverInit(const char *dn, int base_addr);
static asynStatus getAddr(drvPvt_t * pdrvPvt, asynUser * pasynUser,
			      int *paddr, int portOK);
/* asynCommon methods */
static void report(void *drvPvt, FILE * fp, int details);
static asynStatus connect(void *drvPvt, asynUser * pasynUser);
static asynStatus disconnect(void *drvPvt, asynUser * pasynUser);
static asynCommon common = {report, connect, disconnect};
/* asynUInt32Digital methods */
static asynStatus uint32Write(void *drvPvt, asynUser * pasynUser,
			          epicsUInt32 value, epicsUInt32 mask);
static asynStatus uint32Read(void *drvPvt, asynUser * pasynUser,
			         epicsUInt32 * value, epicsUInt32 mask);
/* asynFloat64 methods */
static asynStatus float64Write(void *drvPvt, asynUser * pasynUser, epicsFloat64 value);
static asynStatus float64Read(void *drvPvt, asynUser * pasynUser, epicsFloat64 * value);
static asynStatus create(void *drvPvt, asynUser * pasynUser,
			     const char *drvInfo, const char **pptypeName,
			     size_t * psize)
{
    int i, j;
    drvUserData *pdrvUserData;
    pdrvUserData =
	callocMustSucceed(sizeof(drvUserData) + 1, sizeof(char),
			  "drvUserData create");
    j = sizeof(map) / sizeof(mapType);
    for (i = 0; i < j; i++) {
	if (strcmp(map[i].requestMsg, drvInfo) == 0)
	    break;
    }
    if (i < j) {
	pdrvUserData->type = map[i].type;
	pasynUser->drvUser = (void *) pdrvUserData;
	return asynSuccess;
    } else {
	printf("pas9782Driver,  asynDrvUser interface create(): no such user command %s\n", drvInfo);
	return asynError;
    }
}
static asynStatus getType(void *ddrvPvt, asynUser * pasynUser,
			      const char **pptypeName, size_t * psize)
{
    return asynSuccess;
}
static asynStatus destroy(void *drvPvt, asynUser * pasynUser)
{
    if (pasynUser->drvUser)
	free(pasynUser->drvUser);
    return asynSuccess;
}
int pas9782DriverInit(const char *dn, int base_addr)
{
    drvPvt_t *pdrvPvt;
    char *portName;
    asynStatus status;
    int nbytes;
    asynUInt32Digital *pasynUInt32Digital;
    asynFloat64 *pasynFloat64;
    if (dn == NULL)
	epicsPrintf("Command format: pas9782DriverInit portname, base_addr\n");
    nbytes = sizeof(drvPvt_t) + strlen(dn) + sizeof(asynUInt32Digital) + sizeof(asynFloat64) + 1;
    pdrvPvt =
	callocMustSucceed(nbytes, sizeof(char), "pas9782AsynDriverInit");
    pasynUInt32Digital = (asynUInt32Digital *) (pdrvPvt + 1);
    pasynFloat64 = (asynFloat64 *) (pasynUInt32Digital + 1);
    portName = (char *) (pasynFloat64 + 1);
    strcpy(portName, dn);
    pdrvPvt->portName = portName;
    pdrvPvt->lock = epicsMutexMustCreate();
    pdrvPvt->common.interfaceType = asynCommonType;
    pdrvPvt->common.pinterface = (void *) &common;
    pdrvPvt->common.drvPvt = pdrvPvt;
    pdrvPvt->asynDrvUser.interfaceType = asynDrvUserType;
    pdrvPvt->asynDrvUser.pinterface = (void *) &drvUser;
    pdrvPvt->asynDrvUser.drvPvt = pdrvPvt;
    pasynUInt32Digital->write = uint32Write;
    pasynUInt32Digital->read = uint32Read;
    pdrvPvt->asynUInt32Digital.interfaceType = asynUInt32DigitalType;
    pdrvPvt->asynUInt32Digital.pinterface = pasynUInt32Digital;
    pdrvPvt->asynUInt32Digital.drvPvt = pdrvPvt;
    pasynFloat64->write = float64Write;
    pasynFloat64->read = float64Read;
    pdrvPvt->asynFloat64.interfaceType = asynFloat64Type;
    pdrvPvt->asynFloat64.pinterface = pasynFloat64;
    pdrvPvt->asynFloat64.drvPvt = pdrvPvt;
    if (devRegisterAddress("pas board", atVMEA16, (size_t) base_addr, sizeof(PAS9782_t), (void *) &pdrvPvt->pboard_register)) {
	epicsPrintf(" Addressing Error in pas9782DriverInit for control register.\n");
	return -1;
    }
    epicsPrintf(" pboard_register addr is %p\n", pdrvPvt->pboard_register);
    status =
	pasynManager->registerPort(portName, ASYN_MULTIDEVICE | ASYN_CANBLOCK,
				   1, 100, 0);
    if (status != asynSuccess) {
	epicsPrintf("pas9782DriverInit register Port failed\n");
	return -1;
    }
    status = pasynManager->registerInterface(portName, &pdrvPvt->common);
    if (status != asynSuccess) {
	epicsPrintf("pas9782DriverInit registerInterface  asynCommon failed\n");
	return -1;
    }
    status = pasynManager->registerInterface(portName, &pdrvPvt->asynDrvUser);
    if (status != asynSuccess) {
	epicsPrintf("pas9782DriverInit registerInterface asynDrvUser failed\n");
	return -1;
    }
    status = pasynUInt32DigitalBase->initialize(pdrvPvt->portName,
						&pdrvPvt->asynUInt32Digital);
    if (status != asynSuccess) {
	epicsPrintf("pas9782DriverInit pasynUInt32DigitalBase initialize failed\n");
	return -1;
    }
    status = pasynFloat64Base->initialize(pdrvPvt->portName, &pdrvPvt->asynFloat64);
    if (status != asynSuccess) {
	epicsPrintf("pas9782DriverInit pasynFloat64Base initialize failed\n");
	return -1;
    }
    return 0;
}
static asynStatus getAddr(drvPvt_t * pdrvPvt, asynUser * pasynUser,
			      int *paddr, int portOK)
{
    asynStatus status;
    status = pasynManager->getAddr(pasynUser, paddr);
    if (status != asynSuccess)
	return status;
    if (*paddr >= -1 && *paddr <= NCHANNELS)
	return asynSuccess;
    if (!portOK && *paddr >= 0)
	return asynSuccess;
    epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		  "%s addr %d is illegal; Must be >= %d and <= %d\n",
		  pdrvPvt->portName, *paddr, (portOK ? -1 : 0), NCHANNELS);
    return asynError;
}
/* asynCommon methods */
static void report(void *pvt, FILE * fp, int details)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    fprintf(fp, "pas9782Driver: connected:%s \n",
	    (pdrvPvt->connected ? "Yes" : "No"));
}
static asynStatus connect(void *pvt, asynUser * pasynUser)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    int addr;
    short shval;
    asynStatus status;
    status = getAddr(pdrvPvt, pasynUser, &addr, 1);
    if (status != asynSuccess)
	return status;
    asynPrint(pasynUser, ASYN_TRACE_ERROR,
	      "%s pas9782Driver:connect addr %d\n", pdrvPvt->portName, addr);
    if (addr <= -1) {
	if (pdrvPvt->connected) {
	    asynPrint(pasynUser, ASYN_TRACE_ERROR,
		      "%s pas9782Driver:connect port already connected\n",
		      pdrvPvt->portName);
	    return asynError;
	}
	if (devReadProbe(sizeof(short), (void *) pdrvPvt->pboard_register, (void *) &shval) == 0) {
	    /* initialize board */
	    /* printf("again  pboard_register->csr addr is %p, value is
	     * %04x\n",pdrvPvt->pboard_register,
	     * pdrvPvt->pboard_register->csr); */
	    pdrvPvt->pboard_register->csr = 0x0003;
	    pdrvPvt->connected = 1;
	    pasynManager->exceptionConnect(pasynUser);
	    printf("devReadProbe() OK!\n");
	    return asynSuccess;
	} else {
	    asynPrint(pasynUser, ASYN_TRACE_ERROR,
		      "%s pas9782Driver:can't connect hardware, addr is %d\n", pdrvPvt->portName, addr);
	    return asynError;
	}
    }
    if (addr > NCHANNELS) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:connect:  illegal address %d\n",
		  pdrvPvt->portName, addr);
	return asynError;
    }
    if (pdrvPvt->channel_connected[addr]) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:connect channel %d  already connected\n",
		  pdrvPvt->portName, addr);
	return asynError;
    }
    pdrvPvt->channel_connected[addr] = 1;
    pasynManager->exceptionConnect(pasynUser);
/* 		  pdrvPvt->connected = 1; */
    return asynSuccess;
}
static asynStatus disconnect(void *pvt, asynUser * pasynUser)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    int addr;
    asynStatus status;
    status = getAddr(pdrvPvt, pasynUser, &addr, 1);
    if (status != asynSuccess)
	return status;
    asynPrint(pasynUser, ASYN_TRACE_FLOW,
	      "%s pas9782Driver:disconnect addr %d\n", pdrvPvt->portName,
	      addr);
    if (addr <= -1) {
	if (!pdrvPvt->connected) {
	    asynPrint(pasynUser, ASYN_TRACE_ERROR,
		      "%s pas9782Driver:disconnect port not connected\n",
		      pdrvPvt->portName);
	    return asynError;
	}
	pdrvPvt->connected = 0;
	pasynManager->exceptionDisconnect(pasynUser);
	return asynSuccess;
    }
    if (addr > NCHANNELS) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:connect:  illegal address %d\n",
		  pdrvPvt->portName, addr);
	return asynError;
    }
    if (!pdrvPvt->channel_connected[addr]) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:connect channel %d not  connected\n",
		  pdrvPvt->portName, addr);
	return asynError;
    }
    pdrvPvt->channel_connected[addr] = 0;
    pasynManager->exceptionDisconnect(pasynUser);
    return asynSuccess;
}
static asynStatus uint32Write(void *pvt, asynUser * pasynUser, epicsUInt32 value, epicsUInt32 mask)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    drvUserData *pdrvUserData;
    int addr;
    asynStatus status;
    status = getAddr(pdrvPvt, pasynUser, &addr, 0);
    if (status != asynSuccess)
	return status;
    asynPrint(pasynUser, ASYN_TRACE_FLOW,
	      "%s pas9782Driver: uint32Write value %d\n", pdrvPvt->portName,
	      value);
    if (!pdrvPvt->connected) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:write port not connected\n",
		  pdrvPvt->portName);
	epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		      "%s pas9782Driver:write port not connected\n",
		      pdrvPvt->portName);
	return asynError;
    }
    if (!pdrvPvt->channel_connected[addr]) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:write channel is not connected\n",
		  pdrvPvt->portName);
	epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		      "%s pas9782Driver:write channel is not connected\n",
		      pdrvPvt->portName);
	return asynError;
    }
    epicsMutexMustLock(pdrvPvt->lock);
    pdrvUserData = (drvUserData *) pasynUser->drvUser;
    switch (pdrvUserData->type) {
    case RESET:
	pdrvPvt->pboard_register->csr = pdrvPvt->pboard_register->csr | 0x0010;	/* reset */
	{
	    int i, j;
	    for (i = 0; i < 100; i++)
		j = 0;
	}
	pdrvPvt->pboard_register->csr = 0x0003;
	break;
    case ENABLE:
	if (value == 0)
	    pdrvPvt->pboard_register->csr = pdrvPvt->pboard_register->csr & 0xfffb;	/* 2Mhz disable */
	else
	    pdrvPvt->pboard_register->csr = pdrvPvt->pboard_register->csr | 0x0004;	/* 2Mhz enable */
	break;
    case AMPLITUDE:
	if (value >= 0 && value <= 3)
	    pdrvPvt->pboard_register->ccr[addr - 1] = (pdrvPvt->pboard_register->ccr[addr - 1] & 0xfffc) | value;
	break;
    case FREQ_SELECT:
	pdrvPvt->pboard_register->ccr[addr - 1] = (pdrvPvt->pboard_register->ccr[addr - 1] & 0xfffb) | (value << 2);
	break;
    case PHASE_SELECT:
	pdrvPvt->pboard_register->ccr[addr - 1] = (pdrvPvt->pboard_register->ccr[addr - 1] & 0xfff7) | (value << 3);
	break;
    default:
	break;
    }
    epicsMutexUnlock(pdrvPvt->lock);
    asynPrint(pasynUser, ASYN_TRACEIO_DRIVER,
	      "%s addr %d write %d\n", pdrvPvt->portName, addr, value);
    return asynSuccess;
}
static asynStatus uint32Read(void *pvt, asynUser * pasynUser,
			         epicsUInt32 * value, epicsUInt32 mask)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    drvUserData *pdrvUserData;
    int addr;
    asynStatus status;
    status = getAddr(pdrvPvt, pasynUser, &addr, 0);
    if (status != asynSuccess)
	return status;
    asynPrint(pasynUser, ASYN_TRACE_FLOW,
	      "%s pas9782Driver:uint32Read  value %d\n", pdrvPvt->portName,
	      value);
    if (!pdrvPvt->connected) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:read port  not connected\n",
		  pdrvPvt->portName);
	epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		      "%s pas9782Driver:read port not connected\n",
		      pdrvPvt->portName);
	return asynError;
    }
    if (!pdrvPvt->channel_connected[addr]) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:read channel is not connected\n",
		  pdrvPvt->portName);
	epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		      "%s pas9782Driver:read  channel is not connected\n",
		      pdrvPvt->portName);
	return asynError;
    }
    epicsMutexMustLock(pdrvPvt->lock);
    pdrvUserData = (drvUserData *) pasynUser->drvUser;
    switch (pdrvUserData->type) {
    case RESET:
	*value = ((pdrvPvt->pboard_register->csr & 0x0010) == 0) ? 0 : 1;	/* reset */
	break;
    case ENABLE:
	*value = ((pdrvPvt->pboard_register->csr & 0x0004) == 0) ? 0 : 1;	/* 2Mhz enable */
	break;
    case AMPLITUDE:
	*value = pdrvPvt->pboard_register->ccr[addr - 1] & 0x0003;
	break;
    case FREQ_SELECT:
	*value = ((pdrvPvt->pboard_register->ccr[addr - 1] & 0x0004) == 0) ? 0 : 1;
	break;
    case PHASE_SELECT:
	*value = ((pdrvPvt->pboard_register->ccr[addr - 1] & 0x0008) == 0) ? 0 : 1;
	break;
    case CARD_STATUS:
	*value = ((pdrvPvt->pboard_register->csr & 0x0002) == 0) ? 0 : 1;	/* CARD PASS */
	break;
    default:
	break;
    }
    epicsMutexUnlock(pdrvPvt->lock);
    asynPrint(pasynUser, ASYN_TRACEIO_DRIVER,
	      "%s read %d\n", pdrvPvt->portName, *value);
    return asynSuccess;
}
static asynStatus float64Write(void *pvt, asynUser * pasynUser,
			           epicsFloat64 value)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    epicsFloat64 tmp;
    unsigned long reg;
    int addr;
    asynStatus status;
    drvUserData *pdrvUserData;
    status = getAddr(pdrvPvt, pasynUser, &addr, 0);
    if (status != asynSuccess)
	return status;
    asynPrint(pasynUser, ASYN_TRACE_FLOW,
	      "%s pas9782Driver: float64Write  value %lf\n", pdrvPvt->portName,
	      value);
    if (!pdrvPvt->connected) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:write port not connected\n",
		  pdrvPvt->portName);
	epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		      "%s pas9782Driver:write port not connected\n",
		      pdrvPvt->portName);
	return asynError;
    }
    if (!pdrvPvt->channel_connected[addr]) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s pas9782Driver:write channel is not connected\n",
		  pdrvPvt->portName);
	epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		      "%s pas9782Driver:write channel is not connected\n",
		      pdrvPvt->portName);
	return asynError;
    }
    epicsMutexMustLock(pdrvPvt->lock);
    pdrvUserData = (drvUserData *) pasynUser->drvUser;
    switch (pdrvUserData->type) {
    case FREQUENCY0:
	tmp = (pow(2.0, 32.0) * value) / (2 * pow(10.0, 6.0));
	reg = (unsigned long) tmp;
	pdrvPvt->pboard_register->channel[addr - 1].freq0[0] = (unsigned short) (reg & 0x0000ffff);
	pdrvPvt->pboard_register->channel[addr - 1].freq0[1] = (unsigned short) (reg >> 16);
	break;
    case FREQUENCY1:
	tmp = (pow(2.0, 32.0) * value) / (2 * pow(10.0, 6.0));
	reg = (unsigned long) tmp;
	pdrvPvt->pboard_register->channel[addr - 1].freq1[0] = (unsigned short) (reg & 0x0000ffff);
	pdrvPvt->pboard_register->channel[addr - 1].freq1[1] = (unsigned short) (reg >> 16);
	break;
    case PHASE0:
	pdrvPvt->pboard_register->channel[addr - 1].phase[0] = (unsigned short) (4096 * value / 360.0);
	/* printf(" PHASE0  value is %d\n", (unsigned
	 * short)(4096*value/360.0)); */
	break;
    case PHASE1:
	pdrvPvt->pboard_register->channel[addr - 1].phase[1] = (unsigned short) (4096 * value / 360.0);
	break;
    default:
	break;
    }
    epicsMutexUnlock(pdrvPvt->lock);
    return asynSuccess;
}
static asynStatus float64Read(void *pvt, asynUser * pasynUser,
			          epicsFloat64 * value)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    int addr;
    asynStatus status;
    drvUserData *pdrvUserData;
    status = getAddr(pdrvPvt, pasynUser, &addr, 0);
    if (status != asynSuccess)
	return status;
    asynPrint(pasynUser, ASYN_TRACE_FLOW,
	      "%s %d :float64Read\n", pdrvPvt->portName, addr);
    if (!pdrvPvt->connected) {
	asynPrint(pasynUser, ASYN_TRACE_ERROR,
		  "%s :read  not connected\n", pdrvPvt->portName);
	epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		      "%s :read not connected\n", pdrvPvt->portName);
	return asynError;
    }
    epicsMutexMustLock(pdrvPvt->lock);
    pdrvUserData = (drvUserData *) pasynUser->drvUser;
    switch (pdrvUserData->type) {
    case FREQUENCY0:
	break;
    case PHASE0:
	break;
    default:
	break;
    }
    epicsMutexUnlock(pdrvPvt->lock);
    asynPrint(pasynUser, ASYN_TRACEIO_DRIVER,
	      "%s read %d\n", pdrvPvt->portName, value);
    epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
		  "%s addr These registers are write-only !!\n",
		  pdrvPvt->portName);
    return asynError;
}
/* register pas9782DriverInit*/
static const iocshArg pas9782DriverInitArg0 = {"portName", iocshArgString};
static const iocshArg pas9782DriverInitArg1 =
{"base_addr", iocshArgInt};
static const iocshArg *const pas9782DriverInitArgs[2] =
{&pas9782DriverInitArg0, &pas9782DriverInitArg1
};
static const iocshFuncDef pas9782DriverInitFuncDef =
{"pas9782DriverInit", 2, pas9782DriverInitArgs};
static void pas9782DriverInitCallFunc(const iocshArgBuf * args)
{
    pas9782DriverInit(args[0].sval, args[1].ival);
}
static void pas9782DriverRegister(void)
{
    static int firstTime = 1;
    if (firstTime) {
	firstTime = 0;
	iocshRegister(&pas9782DriverInitFuncDef, pas9782DriverInitCallFunc);
    }
}
epicsExportRegistrar(pas9782DriverRegister);
